source("time.r")


# CREATE COVARIANCE MATRIX.  The covariates for two sets of cases are given by
# vectors x1 and x2.  The parameters of the covariance function are s and r.

gp_cov <- function (x1, x2, s, r)
{
  n <- length(x1)
  m <- length(x2)
  C <- matrix(nrow=n,ncol=m)

  for (j in seq_len(m))
  { C[,j] <- s^2 * exp(-(r*(x1-x2[j]))^2)
  }

  C
}


# LOG LIKELIHOOD FOR GP PARAMETERS.  The covariates are in the vector x,
# the responses in the vector y, and s, r, and v are the parameters.

gp_log_likelihood <- function (x, y, s, r, v)
{
  C <- gp_cov(x,x,s,r)
  diag(C) <- diag(C) + v^2
  U <- chol(C)
  u <- backsolve(U,y,transpose=TRUE)
  as.vector (-sum(u^2)/2 - sum(log(diag(U))) - length(y)*log(2*pi)/2)
}


# SEARCH FOR BEST PARAMETERS.

gp_search <- function (x, y, s.vec, r.vec, v.vec)
{
  best.ll <- -Inf
  for (s in s.vec)
  { for (r in r.vec)
    { for (v in v.vec)
      { ll <- gp_log_likelihood(x,y,s,r,v)
        if (ll >= best.ll)
        { best.s <- s
          best.r <- r
          best.v <- v
          best.ll <- ll
        }
      }
    }
  }
  c(s=best.s,r=best.r,v=best.v)
}


# PREDICT RESPONSES IN TEST CASES.

gp_predict <- function (x, y, s, r, v, x.test)
{
  C <- gp_cov(x,x,s,r)
  diag(C) <- diag(C) + v^2
  U <- chol(C)
  u <- backsolve(U,y,transpose=TRUE)
  u <- backsolve(U,u)
  K <- gp_cov(x,x.test,s,r)
  as.vector (t(K) %*% u)
}

test.name <- "prg-gp"

tf <- function (x) 1.2 * sin (0.3 + 0.2*x^2 + 2.7*sin(2*x+0.2))

set.seed(1)

f <- function (N, n)
{ 
  set.seed(1)

  hlist <<- list()

  for (i in 1:N)
  {
    x <- rnorm(n)
    x.test <- seq(-2,2,length=1000)

    y <- tf(x) + rnorm(n,0,0.11)

    h <- gp_search (x, y, s.vec = c(0.7,1.0,1.2,1.4,2.0), 
                          r.vec = c(1.0,1.4,2.0,2.8), 
                          v.vec = c(0.1,0.14,0.2,0.28))
    hlist[[i]] <<- h

    p <- gp_predict (x, y, h["s"], h["r"], h["v"], x.test)
  }

  cbind (x=x.test, y=tf(x.test), p=p)
}

sys.time(R<-f(10,100))

print(hlist)
print(R[c(1,500,1000),"p"])

sys.time(R<-f(1,350))

print(R[c(1,500,1000),"p"])

# Uncomment to check result...
#
# plot(R[,"x"],R[,"y"],type="l")
# points(R[,"x"],R[,"p"],pch=20)
